from datetime import datetime
from dcicutils import ff_utils
from dcicutils.s3_utils import s3Utils
from .helpers import wfr_utils
from .helpers import wfrset_utils

# Use confchecks to import decorators object and its methods for each check module
# rather than importing check_function, action_function, CheckResult, ActionResult
# individually - they're now part of class Decorators in foursight-core::decorators
# that requires initialization with foursight prefix.
from .helpers.confchecks import *

lambda_limit = wfr_utils.lambda_limit


@check_function(lab_title=None, start_date=None, action="chipseq_start")
def chipseq_status(connection, **kwargs):
    """
    Keyword arguments:
    lab_title -- limit search with a lab i.e. Bing+Ren, UCSD
    start_date -- limit search to files generated since a date formatted YYYY-MM-DD
    run_time -- assume runs beyond run_time are dead
    """
    start = datetime.utcnow()
    check = CheckResult(connection, 'chipseq_status')
    my_auth = connection.ff_keys
    check.action = "chipseq_start"
    check.description = "run missing steps and add processing results to processed files, match set status"
    check.brief_output = []
    check.summary = ""
    check.full_output = {'skipped': [], 'running_runs': [], 'needs_runs': [],
                         'completed_runs': [], 'problematic_runs': []}
    check.status = 'PASS'
    exp_type = 'ChIP-seq'
    # completion tag
    tag = wfr_utils.accepted_versions[exp_type][-1]
    # check indexing queue
    check, skip = wfr_utils.check_indexing(check, connection)
    if skip:
        return check
    # Build the query, add date and lab if available
    query = wfr_utils.build_exp_type_query(exp_type, kwargs)
    res = ff_utils.search_metadata(query, key=my_auth)
    print(len(res))

    if not res:
        check.summary = 'All Good!'
        return check
    # run step 0 on all experiments with more than 2 sets of files
    # for control sets, run step1c on each experiment and finish
    # for non-control sets, run step1 on each experiment, check if control is ready, run step2 on set
    step0_name = 'merge-fastq'
    step1_name = 'encode-chipseq-aln-chip'
    step1c_name = 'encode-chipseq-aln-ctl'
    step2_name = 'encode-chipseq-postaln'

    for a_set in res:
        set_acc = a_set['accession']
        all_items, all_uuids = ff_utils.expand_es_metadata([a_set['uuid']], my_auth,
                                                           store_frame='embedded',
                                                           add_pc_wfr=True,
                                                           ignore_field=[  # 'experiment_relation',
                                                                         'biosample_relation',
                                                                         'references',
                                                                         'reference_pubs'])
        now = datetime.utcnow()
        print(a_set['accession'], (now-start).seconds, len(all_uuids))
        if (now-start).seconds > lambda_limit:
            break
        # are all files uploaded ?
        all_uploaded = True
        for a_file in all_items['file_fastq']:
            if a_file['status'] in ['uploading', 'upload failed']:
                all_uploaded = False

        if not all_uploaded:
            final_status = a_set['accession'] + ' skipped, waiting for file upload'
            print(final_status)
            check.brief_output.append(final_status)
            check.full_output['skipped'].append({a_set['accession']: 'files status uploading'})
            continue

        all_wfrs = all_items.get('workflow_run_awsem', []) + all_items.get('workflow_run_sbg', [])
        all_files = [i for typ in all_items for i in all_items[typ] if typ.startswith('file_')]
        all_qcs = [i for typ in all_items for i in all_items[typ] if typ.startswith('quality_metric')]
        library = {'wfrs': all_wfrs, 'files': all_files, 'qcs': all_qcs}
        keep = {'missing_run': [], 'running': [], 'problematic_run': []}
        # if all completed, patch this info
        complete = {'patch_opf': [],
                    'add_tag': []}
        set_acc = a_set['accession']

        # some feature to extract from each set
        control = ""  # True or False (True if set is control)
        control_set = ""  # None if there are no control experiments or if the set is control
        target_type = ""  # Histone or TF (or None for control)
        paired = ""  # single or paired , checked for each experiment
        organism = ""
        replicate_exps = a_set['replicate_exps']
        replicate_exps = sorted(replicate_exps, key=lambda x: [x['bio_rep_no'], x['tec_rep_no']])
        # get organism, target and control from the first replicate
        f_exp = replicate_exps[0]['replicate_exp']['uuid']
        # have to do another get for control experiments if there is one
        f_exp_resp = [i for i in all_items['experiment_seq'] if i['uuid'] == f_exp][0]
        control, control_set, target_type, organism = wfr_utils.get_chip_info(f_exp_resp, all_items)
        print('ORG:', organism, "CONT:", control, "TARGET:", target_type, "CONT_SET:", control_set)
        set_summary = " - ".join([set_acc, str(organism), str(target_type), str(control)])
        # sanity checks
        # if control and also has an AB with target
        if control and target_type:
            set_summary += "| error - has target and is control"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: set_summary})
            continue
        # can only process mouse and human at the moment
        if organism not in ['mouse', 'human']:
            set_summary += "| organism not ready for chip"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: set_summary})
            continue
        # if not control, we need a target
        if not control and not target_type:
            set_summary += "| missing target type"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: set_summary})
            continue
        # collect results from step1 runs for step2
        ta = []
        taxcor = []
        ta_cnt = []
        # track if all experiments completed step0 and step1
        ready_for_step2 = True
        for an_exp in replicate_exps:
            # track if all experiments completed step0
            ready_for_step1 = True
            # track if all control experiments are completed processing
            control_ready = True
            exp_id = an_exp['replicate_exp']['accession']
            exp_resp = [i for i in all_items['experiment_seq'] if i['accession'] == exp_id][0]
            exp_files, paired = wfr_utils.get_chip_files(exp_resp, all_files)
            # if there are more then 2 files, we need to merge:
            print(exp_id, len(exp_files), paired)
            # if too many input, merge them
            if len(exp_files) > 2:
                # exp_files format [[pair1,pair2], [pair1, pair2]]  @id
                input_list = []
                if paired == 'paired':
                    # first add paired end 1s
                    input_list.append([i[0] for i in exp_files])
                    input_list.append([i[1] for i in exp_files])
                elif paired == 'single':
                    input_list.append([i[0] for i in exp_files])
                # collect files for step1 and step1c
                merged_files = []
                step0_status = 'complete'
                merge_enum = 0
                # if paired, need to run merge twice for each end
                for merge_case in input_list:
                    merge_enum += 1
                    # RUN STEP 0
                    s0_input_files = {'input_fastqs': merge_case}
                    s0_tag = exp_id + '_p' + str(merge_enum)
                    keep, step0_status, step0_output = wfr_utils.stepper(library, keep,
                                                                         'step0', s0_tag, merge_case,
                                                                         s0_input_files, step0_name, 'merged_fastq', organism=organism)
                    if step0_status == 'complete':
                        merged_files.append(step0_output)
                    else:
                        ready_for_step1 = False

                if ready_for_step1:
                    # rewrite exp_files with merged ones
                    exp_files = [[]]
                    for a_merged in merged_files:
                        exp_files[0].append(a_merged)
            # if step0 was not complete, skip checks for step2
            if not ready_for_step1:
                ready_for_step2 = False
                continue

            # step1 references:
            input_files = {}
            if organism == 'human':
                org = 'hs'
                input_files['chip.bwa_idx_tar'] = '/files-reference/4DNFIZQB369V/'
                input_files['chip.blacklist'] = '/files-reference/4DNFIZ1TGJZR/'
                input_files['chip.chrsz'] = '/files-reference/4DNFIZJB62D1/'
                input_files['additional_file_parameters'] = {"chip.bwa_idx_tar": {"rename": "GRCh38_no_alt_analysis_set_GCA_000001405.15.fasta.tar"}}
            if organism == 'mouse':
                org = 'mm'
                input_files['chip.bwa_idx_tar'] = '/files-reference/4DNFIZ2PWCC2/'
                input_files['chip.blacklist'] = '/files-reference/4DNFIZ3FBPK8/'
                input_files['chip.chrsz'] = '/files-reference/4DNFIBP173GC/'
                input_files['additional_file_parameters'] = {"chip.bwa_idx_tar": {"rename": "mm10_no_alt_analysis_set_ENCODE.fasta.tar"}}
            # step1 Parameters
            parameters = {}
            parameters["chip.gensz"] = org
            if paired == 'single':
                frag_temp = [300]
                fraglist = frag_temp * len(exp_files)
                parameters['chip.fraglen'] = fraglist
                parameters['chip.paired_end'] = False
            elif paired == 'paired':
                parameters['chip.paired_end'] = True

            # run step1 for control
            if control:
                # control run on tf mode
                # input_files = {'chip.ctl_fastqs': [exp_files]}
                input_files['chip.ctl_fastqs'] = [exp_files]
                control_parameters = {
                    "chip.pipeline_type": 'tf',
                    "chip.choose_ctl.always_use_pooled_ctl": True,
                    "chip.bam2ta_ctl.regex_grep_v_ta": "chr[MUE]|random|alt",
                    "chip.bwa_ctl.cpu": 8,
                    "chip.merge_fastq_ctl.cpu": 8,
                    "chip.filter_ctl.cpu": 8,
                    "chip.bam2ta_ctl.cpu": 8,
                    "chip.align_only": True
                }
                parameters.update(control_parameters)

                s1c_input_files = input_files
                s1c_tag = exp_id
                keep, step1c_status, step1c_output = wfr_utils.stepper(library, keep,
                                                                       'step1c', s1c_tag, exp_files,
                                                                       s1c_input_files, step1c_name, 'chip.first_ta_ctl',
                                                                       additional_input={'parameters': parameters}, organism=organism)
                if step1c_status == 'complete':
                    # accumulate files to patch on experiment
                    patch_data = [step1c_output, ]
                    complete['patch_opf'].append([exp_id, patch_data])
                else:
                    # don't patch anything if at least one exp is still missing
                    ready_for_step2 = False
                print('step1c')
                print(step1c_status, step1c_output)

            # run step1
            else:
                # input_files = {'chip.fastqs': [exp_files]}
                input_files['chip.fastqs'] = [exp_files]
                exp_parameters = {
                    "chip.pipeline_type": target_type,
                    "chip.choose_ctl.always_use_pooled_ctl": True,
                    "chip.bam2ta.regex_grep_v_ta": "chr[MUE]|random|alt",
                    "chip.bwa.cpu": 8,
                    "chip.merge_fastq.cpu": 8,
                    "chip.filter.cpu": 8,
                    "chip.bam2ta.cpu": 8,
                    "chip.xcor.cpu": 8,
                    "chip.align_only": True
                }
                parameters.update(exp_parameters)

                s1_input_files = input_files
                s1_tag = exp_id
                # if complete, step1_output will have a list of 2 files, first_ta, and fist_ta_xcor
                keep, step1_status, step1_output = wfr_utils.stepper(library, keep,
                                                                     'step1', s1_tag, exp_files,
                                                                     s1_input_files, step1_name, ['chip.first_ta', 'chip.first_ta_xcor'],
                                                                     additional_input={'parameters': parameters}, organism=organism)
                if step1_status == 'complete':
                    exp_ta_file = step1_output[0]
                    exp_taxcor_file = step1_output[1]
                    # accumulate files to patch on experiment
                    patch_data = [exp_ta_file, ]
                    complete['patch_opf'].append([exp_id, patch_data])
                    ta.append(exp_ta_file)
                    taxcor.append(exp_taxcor_file)

                    # find the control file if there is a control set found
                    if control_set:
                        try:
                            exp_cnt_ids = [i['experiment'] for i in exp_resp['experiment_relation'] if i['relationship_type'] == 'controlled by']
                            exp_cnt_ids = [i['@id'] for i in exp_cnt_ids]
                        except:
                            control_ready = False
                            print('Control Relation has problems for this exp', exp_id)
                            continue
                        if len(exp_cnt_ids) != 1:
                            control_ready = False
                            print('Multiple controls for this exp', exp_id)
                            continue
                        exp_cnt_id = exp_cnt_ids[0]
                        print('controled by set', exp_cnt_id)
                        # have to do a get for the control experiment
                        exp_cnt_resp = [i for i in all_items['experiment_seq'] if i['@id'] == exp_cnt_id][0]
                        cont_file = ''
                        # check opf for control file
                        for opf_case in exp_cnt_resp.get('other_processed_files', []):
                            if opf_case['title'] == 'ENCODE ChIP-Seq Pipeline - Preliminary Files':
                                opf_files = opf_case['files']
                                assert len(opf_files) == 1
                                cont_file = opf_files[0]['@id']
                        # if not in opf, check processed files
                        if not cont_file:
                            pf_list = exp_cnt_resp.get('processed_files', [])
                            if pf_list:
                                if pf_list:
                                    assert len(pf_list) == 1
                                    cont_file = pf_list[0]['@id']
                        # did we find it, if so, add it to ta_cnt
                        if cont_file:
                            ta_cnt.append(cont_file)
                        else:
                            control_ready = False

                else:
                    # don't patch anything if at least one exp is still missing
                    ready_for_step2 = False
                print('step1')
                print(step1_status, step1_output, control_ready)
        # back to set level
        final_status = set_acc  # start the reporting with acc
        all_completed = False
        # is step0 step1 complete
        if ready_for_step2 and not control_ready:
            final_status += ' waiting for control experiments to finish processing'
        elif ready_for_step2:
            # for control, add tag to set, and files to experiments
            if control:
                complete['add_tag'] = [set_acc, tag]
            # for non controls check for step2
            else:
                # this only works with 2 experiments, if 3, pick best 2, if more, skip for now
                if len(ta) > 3:
                    set_summary += "| skipped - more then 3 experiments in set, can not process at the moment"
                    check.brief_output.append(set_summary)
                    check.full_output['skipped'].append({set_acc: set_summary})
                    continue
                if len(ta) > 2:
                    ta_2 = []
                    taxcor_2 = []
                    print('ExperimentSet has 3 experiments, selecting best 2')
                    ta_2 = wfr_utils.select_best_2(ta, all_files, all_qcs)
                    # xcor does not have qc, use ta indexes to find the correct files
                    for ta_f in ta_2:
                        taxcor_2.append(taxcor[ta.index(ta_f)])
                    ta = ta_2
                    taxcor = taxcor_2
                    # for control files ,also select best2
                    ta_cnt = wfr_utils.select_best_2(ta_cnt, all_files, all_qcs)

                # collect step2 input files
                s2_input_files = {}
                if organism == 'human':
                    org = 'hs'
                    s2_input_files['chip.blacklist'] = '/files-reference/4DNFIZ1TGJZR/'
                    s2_input_files['chip.chrsz'] = '/files-reference/4DNFIZJB62D1/'
                if organism == 'mouse':
                    org = 'mm'
                    s2_input_files['chip.blacklist'] = '/files-reference/4DNFIZ3FBPK8/'
                    s2_input_files['chip.chrsz'] = '/files-reference/4DNFIBP173GC/'

                def rename_chip(input_at_id_list):
                    # rename bed.gz to tagAlign.gz
                    renamed = []
                    for a_file in input_at_id_list:
                        acc = a_file.split('/')[2]
                        renamed.append(acc + '.tagAlign.gz')
                    return renamed

                s2_input_files['additional_file_parameters'] = {}
                s2_input_files['chip.tas'] = ta
                s2_input_files['additional_file_parameters']['chip.tas'] = {"rename": rename_chip(ta)}
                s2_input_files['chip.bam2ta_no_filt_R1.ta'] = taxcor
                s2_input_files['additional_file_parameters']['chip.bam2ta_no_filt_R1.ta'] = {"rename": rename_chip(taxcor)}
                if ta_cnt:
                    s2_input_files['chip.ctl_tas'] = ta_cnt
                    s2_input_files['additional_file_parameters']['chip.ctl_tas'] = {"rename": rename_chip(ta_cnt)}

                # collect parameters
                parameters = {}
                if paired == 'single':
                    chip_p = False
                elif paired == 'paired':
                    chip_p = True
                if not control_set:
                    if target_type == 'histone':
                        set_summary += "| skipped - histone without control needs attention, ie change to tf"
                        check.brief_output.append(set_summary)
                        check.full_output['skipped'].append({set_acc: set_summary})
                        continue
                run_ids = {'desc': set_acc + a_set.get('description', '')}
                parameters = {
                    "chip.pipeline_type": target_type,
                    "chip.paired_end": chip_p,
                    "chip.choose_ctl.always_use_pooled_ctl": True,
                    "chip.qc_report.desc": run_ids['desc'],
                    "chip.gensz": org,
                    "chip.xcor.cpu": 4,
                }
                if paired == 'single':
                    frag_temp = [300]
                    fraglist = frag_temp * len(ta)
                    parameters['chip.fraglen'] = fraglist

                # if the target is a tf and there is no control, use macs2
                if not control_set:
                    if target_type == 'tf':
                        parameters['chip.peak_caller'] = "macs2"

                s2_tag = set_acc
                # if complete, step1_output will have a list of 2 files, first_ta, and fist_ta_xcor
                keep, step2_status, step2_output = wfr_utils.stepper(library, keep,
                                                                     'step2', s2_tag, ta,
                                                                     s2_input_files, step2_name,
                                                                     ['chip.optimal_peak', 'chip.conservative_peak', 'chip.sig_fc'],
                                                                     additional_input={'parameters': parameters}, organism=organism)
                if step2_status == 'complete':
                    set_opt_peak = step2_output[0]
                    set_cons_peak = step2_output[1]
                    set_sig_fc = step2_output[2]
                    # accumulate files to patch on experiment
                    patch_data = [set_opt_peak, set_cons_peak, set_sig_fc]
                    complete['patch_opf'].append([set_acc, patch_data])
                    complete['add_tag'] = [set_acc, tag]
                    all_completed = True

        # unpack results
        missing_run = keep['missing_run']
        running = keep['running']
        problematic_run = keep['problematic_run']
        if all_completed:
            final_status += ' completed'
        else:
            if missing_run:
                final_status += ' |Missing: ' + " ".join([i[0] for i in missing_run])
            if running:
                final_status += ' |Running: ' + " ".join([i[0] for i in running])
            if problematic_run:
                final_status += ' |Problem: ' + " ".join([i[0] for i in problematic_run])

        # add dictionaries to main ones
        check.brief_output.append(final_status)
        print(final_status)
        if running:
            check.full_output['running_runs'].append({set_acc: running})
        if missing_run:
            check.full_output['needs_runs'].append({set_acc: missing_run})
        if problematic_run:
            check.full_output['problematic_runs'].append({set_acc: problematic_run})
        # if made it till the end
        if complete.get('add_tag'):
            assert not running
            assert not problematic_run
            assert not missing_run
            check.full_output['completed_runs'].append(complete)

    # complete check values
    check.summary = ""
    if check.full_output['running_runs']:
        check.summary = str(len(check.full_output['running_runs'])) + ' running|'
    if check.full_output['skipped']:
        check.summary += str(len(check.full_output['skipped'])) + ' skipped|'
        check.status = 'WARN'
    if check.full_output['needs_runs']:
        check.summary += str(len(check.full_output['needs_runs'])) + ' missing|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['completed_runs']:
        check.summary += str(len(check.full_output['completed_runs'])) + ' completed|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['problematic_runs']:
        check.summary += str(len(check.full_output['problematic_runs'])) + ' problem|'
        check.status = 'WARN'
    return check


@action_function(start_runs=True, patch_completed=True)
def chipseq_start(connection, **kwargs):
    """Start runs by sending compiled input_json to run_workflow endpoint"""
    start = datetime.utcnow()
    action = ActionResult(connection, 'chipseq_start')
    my_auth = connection.ff_keys
    my_env = connection.ff_env
    fs_env = connection.fs_env
    chipseq_check_result = action.get_associated_check_result(kwargs).get('full_output', {})
    missing_runs = []
    patch_meta = []
    if kwargs.get('start_runs'):
        missing_runs = chipseq_check_result.get('needs_runs')
    if kwargs.get('patch_completed'):
        patch_meta = chipseq_check_result.get('completed_runs')
    action = wfr_utils.start_tasks(missing_runs, patch_meta, action, my_auth, my_env, fs_env, start,  move_to_pc=True, runtype='chip')
    return action


@check_function(lab_title=None, start_date=None, pick_best_2=False, action="atacseq_start")
def atacseq_status(connection, **kwargs):
    """
    Keyword arguments:
    lab_title -- limit search with a lab i.e. Bing+Ren, UCSD
    start_date -- limit search to files generated since a date formatted YYYY-MM-DD
    run_time -- assume runs beyond run_time are dead
    pick_best_2 -- False by default. If set the True, for sets more than 2 experiments,
                   2 best will be used instead of running mergebed
    """
    start = datetime.utcnow()
    check = CheckResult(connection, 'atacseq_status')
    my_auth = connection.ff_keys
    check.action = "atacseq_start"
    check.description = "run missing steps and add processing results to processed files, match set status"
    check.brief_output = []
    check.summary = ""
    check.full_output = {'skipped': [], 'running_runs': [], 'needs_runs': [],
                         'completed_runs': [], 'problematic_runs': []}
    check.status = 'PASS'
    exp_type = 'ATAC-seq'
    # completion tag
    tag = wfr_utils.accepted_versions[exp_type][-1]
    pick_best_2 = kwargs.get('pick_best_2', False)
    # check indexing queue
    check, skip = wfr_utils.check_indexing(check, connection)
    if skip:
        return check
    # Build the query, add date and lab if available
    query = wfr_utils.build_exp_type_query(exp_type, kwargs)
    res = ff_utils.search_metadata(query, key=my_auth)
    print(len(res))

    if not res:
        check.summary = 'All Good!'
        return check
    # run step 0 on all experiments with more than 2 sets of files
    # step1 on each experiment,if multiple exps, merge beds, run step3 on set
    step0_name = 'merge-fastq'
    step1_name = 'encode-atacseq-aln'
    step2_name = 'mergebed'
    step3_name = 'encode-atacseq-postaln'

    for a_set in res:
        set_acc = a_set['accession']
        all_items, all_uuids = ff_utils.expand_es_metadata([a_set['uuid']], my_auth,
                                                           store_frame='embedded',
                                                           add_pc_wfr=True,
                                                           ignore_field=['experiment_relation',
                                                                         'biosample_relation',
                                                                         'references',
                                                                         'reference_pubs'])
        now = datetime.utcnow()
        print(a_set['accession'], (now-start).seconds, len(all_uuids))
        if (now-start).seconds > lambda_limit:
            break
        # are all files uploaded ?
        all_uploaded = True
        for a_file in all_items['file_fastq']:
            if a_file['status'] in ['uploading', 'upload failed']:
                all_uploaded = False
        if not all_uploaded:
            final_status = a_set['accession'] + ' skipped, waiting for file upload'
            print(final_status)
            check.brief_output.append(final_status)
            check.full_output['skipped'].append({a_set['accession']: 'files status uploading'})
            continue
        all_wfrs = all_items.get('workflow_run_awsem', []) + all_items.get('workflow_run_sbg', [])
        all_files = [i for typ in all_items for i in all_items[typ] if typ.startswith('file_')]
        all_qcs = [i for typ in all_items for i in all_items[typ] if typ.startswith('quality_metric')]
        library = {'wfrs': all_wfrs, 'files': all_files, 'qcs': all_qcs}
        keep = {'missing_run': [], 'running': [], 'problematic_run': []}
        # if all completed, patch this info
        complete = {'patch_opf': [],
                    'add_tag': []}
        set_acc = a_set['accession']

        # some feature to extract from each set
        paired = ""  # single or paired , checked for each experiment
        organism = ""
        replicate_exps = a_set['replicate_exps']
        replicate_exps = sorted(replicate_exps, key=lambda x: [x['bio_rep_no'], x['tec_rep_no']])
        # get organism
        f_exp = replicate_exps[0]['replicate_exp']['uuid']
        # have to do another get for control experiments if there is one
        f_exp_resp = [i for i in all_items['experiment_atacseq'] if i['uuid'] == f_exp][0]
        biosample = f_exp_resp['biosample']
        organism = list(set([bs['organism']['name'] for bs in biosample['biosource']]))[0]
        set_summary = " - ".join([set_acc, str(organism)])
        print(set_summary)
        # sanity checks
        # can only process mouse and human at the moment
        if organism not in ['mouse', 'human']:
            set_summary += "| organism not ready for atac"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: set_summary})
            continue

        # collect results from step1 runs for step2
        ta = []
        # track if all experiments completed step0 and step1
        ready_for_step2 = True
        for an_exp in replicate_exps:
            # track if all experiments completed step0
            ready_for_step1 = True
            exp_id = an_exp['replicate_exp']['accession']
            exp_resp = [i for i in all_items['experiment_atacseq'] if i['accession'] == exp_id][0]
            # exp_files [[pair1,pair2], [pair1, pair2]]
            exp_files, paired = wfr_utils.get_chip_files(exp_resp, all_files)
            # if there are more then 2 files, we need to merge:
            print(exp_id, len(exp_files), paired)
            # if too many input, merge them
            if len(exp_files) > 2:
                # exp_files format [[pair1,pair2], [pair1, pair2]]  @id
                input_list = []
                if paired == 'paired':
                    # first add paired end 1s
                    input_list.append([i[0] for i in exp_files])
                    input_list.append([i[1] for i in exp_files])
                elif paired == 'single':
                    input_list.append([i[0] for i in exp_files])
                # collect files for step1 and step1c
                merged_files = []
                step0_status = 'complete'
                merge_enum = 0
                # if paired, need to run merge twice for each end
                for merge_case in input_list:
                    merge_enum += 1
                    # RUN STEP 0
                    s0_input_files = {'input_fastqs': merge_case}
                    s0_tag = exp_id + '_p' + str(merge_enum)
                    keep, step0_status, step0_output = wfr_utils.stepper(library, keep,
                                                                         'step0', s0_tag, merge_case,
                                                                         s0_input_files, step0_name, 'merged_fastq')
                    if step0_status == 'complete':
                        merged_files.append(step0_output)
                    else:
                        ready_for_step1 = False

                if ready_for_step1:
                    # rewrite exp_files with merged ones
                    exp_files = [[]]
                    for a_merged in merged_files:
                        exp_files[0].append(a_merged)
            # if step0 was not complete, skip checks for step2
            if not ready_for_step1:
                ready_for_step2 = False
                continue

            # step1 files
            # references
            input_files = {}
            if organism == 'human':
                org = 'hs'
                input_files['atac.bowtie2_idx_tar'] = '/files-reference/4DNFIMQPTYDY/'
                input_files['atac.blacklist'] = '/files-reference/4DNFIZ1TGJZR/'
                input_files['atac.chrsz'] = '/files-reference/4DNFIZJB62D1/'
                input_files['additional_file_parameters'] = {"atac.bowtie2_idx_tar": {"rename": "GRCh38_no_alt_analysis_set_GCA_000001405.15.fasta.tar"}}
            if organism == 'mouse':
                org = 'mm'
                input_files['atac.bowtie2_idx_tar'] = '/files-reference/4DNFI2493SDN/'
                input_files['atac.blacklist'] = '/files-reference/4DNFIZ3FBPK8/'
                input_files['atac.chrsz'] = '/files-reference/4DNFIBP173GC/'
                input_files['additional_file_parameters'] = {"atac.bowtie2_idx_tar": {"rename": "mm10_no_alt_analysis_set_ENCODE.fasta.tar"}}
            # add input files
            input_files['atac.fastqs'] = [exp_files]
            # step1 Parameters
            parameters = {
                "atac.pipeline_type": 'atac',
                "atac.gensz": org,
                "atac.bam2ta.regex_grep_v_ta": "chr[MUE]|random|alt",
                "atac.disable_ataqc": True,
                "atac.enable_xcor": False,
                "atac.trim_adapter.auto_detect_adapter": True,
                "atac.bowtie2.cpu": 4,
                "atac.filter.cpu": 4,
                "atac.bam2ta.cpu": 4,
                "atac.trim_adapter.cpu": 4,
                "atac.align_only": True
            }
            if paired == 'single':
                frag_temp = [300]
                fraglist = frag_temp * len(exp_files)
                parameters['atac.fraglen'] = fraglist
                parameters['atac.paired_end'] = False
            elif paired == 'paired':
                parameters['atac.paired_end'] = True

            s1_input_files = input_files
            s1_tag = exp_id
            # if complete, step1_output will have a list of 2 files, first_ta, and fist_ta_xcor
            keep, step1_status, step1_output = wfr_utils.stepper(library, keep,
                                                                 'step1', s1_tag, exp_files,
                                                                 s1_input_files, step1_name, 'atac.first_ta',
                                                                 additional_input={'parameters': parameters})
            if step1_status == 'complete':
                # accumulate files to patch on experiment
                patch_data = [step1_output, ]
                complete['patch_opf'].append([exp_id, patch_data])
                ta.append(step1_output)
            else:
                # don't patch anything if at least one exp is still missing
                ready_for_step2 = False
            print('step1', step1_status, step1_output)

        # back to set level
        final_status = set_acc  # start the reporting with acc
        all_completed = False
        # is step0 step1 complete
        if ready_for_step2:
            # Following was the proposed logic, but it is not implemented
            # Currently, for sets with more than 2 experiments, there are 2 options
            # 1) pick best 2,   2) run mergebed (default)

            # Proposed logic
            # if there are more then 2 experiments, check the number of biological replicates
            # if there is 1 Biological Replicate
            # -pick best 2 exp
            # if there are 2 Biological replicates
            #  - run mergebed on bioreps with more then 1 technical replicate
            # if there are 3 Biological replicates
            # - if there are 3 total experiments (1 in each biological rep), pick best 2
            # - else, run mergebed on bioreps with more then 1 technical replicate, and pick best 2 biorep
            # if there are 4 or more Biolofical replicates
            # - run mergebed on bioreps with more then 1 technical replicate, and pick best 2 biorep
            # this only works with 2 experiments, if 3, pick best 2, if more, skip for now
            ready_for_step3 = True
            if len(ta) > 2:
                if pick_best_2:
                    # pick best 2 - False by default
                    print('ExperimentSet has 3 experiments, selecting best 2')
                    ta = wfr_utils.select_best_2(ta, all_files, all_qcs)
                else:
                    # run mergebed - default option
                    s2_input_files = {'input_bed': ta}
                    s2_tag = set_acc
                    # if complete, step1_output will have a list of 2 files, first_ta, and fist_ta_xcor
                    keep, step2_status, step2_output = wfr_utils.stepper(library, keep,
                                                                         'step2', s2_tag, ta,
                                                                         s2_input_files, step2_name, 'merged_bed')
                    if step2_status == 'complete':
                        ta = [step2_output, ]
                    else:
                        ready_for_step3 = False
            if ready_for_step3:
                # collect step3 input files
                s3_input_files = {}
                if organism == 'human':
                    org = 'hs'
                    s3_input_files['atac.blacklist'] = '/files-reference/4DNFIZ1TGJZR/'
                    s3_input_files['atac.chrsz'] = '/files-reference/4DNFIZJB62D1/'
                if organism == 'mouse':
                    org = 'mm'
                    s3_input_files['atac.blacklist'] = '/files-reference/4DNFIZ3FBPK8/'
                    s3_input_files['atac.chrsz'] = '/files-reference/4DNFIBP173GC/'

                def rename_chip(input_at_id_list):
                    # rename bed.gz to tagAlign.gz
                    renamed = []
                    for a_file in input_at_id_list:
                        acc = a_file.split('/')[2]
                        renamed.append(acc + '.tagAlign.gz')
                    return renamed

                s3_input_files['additional_file_parameters'] = {}
                s3_input_files['atac.tas'] = ta
                s3_input_files['additional_file_parameters']['chip.tas'] = {"rename": rename_chip(ta)}
                # collect parameters
                if paired == 'single':
                    chip_p = False
                elif paired == 'paired':
                    chip_p = True
                parameters = {
                    "atac.pipeline_type": 'atac',
                    "atac.paired_end": chip_p,
                    "atac.gensz": org,
                    "atac.disable_ataqc": True,
                    "atac.enable_xcor": False,
                }
                if paired == 'single':
                    frag_temp = [300]
                    fraglist = frag_temp * len(ta)
                    parameters['atac.fraglen'] = fraglist

                s3_tag = set_acc
                # if complete, step1_output will have a list of 2 files, first_ta, and fist_ta_xcor
                keep, step3_status, step3_output = wfr_utils.stepper(library, keep,
                                                                     'step3', s3_tag, ta,
                                                                     s3_input_files, step3_name,
                                                                     ['atac.optimal_peak', 'atac.conservative_peak', 'atac.sig_fc'],
                                                                     additional_input={'parameters': parameters})
                if step3_status == 'complete':
                    set_opt_peak = step3_output[0]
                    set_cons_peak = step3_output[1]
                    set_sig_fc = step3_output[2]
                    # accumulate files to patch on experiment
                    patch_data = [set_opt_peak, set_cons_peak, set_sig_fc]
                    complete['patch_opf'].append([set_acc, patch_data])
                    complete['add_tag'] = [set_acc, tag]
                    all_completed = True

        # unpack results
        missing_run = keep['missing_run']
        running = keep['running']
        problematic_run = keep['problematic_run']
        if all_completed:
            final_status += ' completed'
        else:
            if missing_run:
                final_status += ' |Missing: ' + " ".join([i[0] for i in missing_run])
            if running:
                final_status += ' |Running: ' + " ".join([i[0] for i in running])
            if problematic_run:
                final_status += ' |Problem: ' + " ".join([i[0] for i in problematic_run])

        # add dictionaries to main ones
        check.brief_output.append(final_status)
        print(final_status)
        if running:
            check.full_output['running_runs'].append({set_acc: running})
        if missing_run:
            check.full_output['needs_runs'].append({set_acc: missing_run})
        if problematic_run:
            check.full_output['problematic_runs'].append({set_acc: problematic_run})
        # if made it till the end
        if complete.get('add_tag'):
            assert not running
            assert not problematic_run
            assert not missing_run
            check.full_output['completed_runs'].append(complete)

    # complete check values
    check.summary = ""
    if check.full_output['running_runs']:
        check.summary = str(len(check.full_output['running_runs'])) + ' running|'
    if check.full_output['skipped']:
        check.summary += str(len(check.full_output['skipped'])) + ' skipped|'
        check.status = 'WARN'
    if check.full_output['needs_runs']:
        check.summary += str(len(check.full_output['needs_runs'])) + ' missing|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['completed_runs']:
        check.summary += str(len(check.full_output['completed_runs'])) + ' completed|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['problematic_runs']:
        check.summary += str(len(check.full_output['problematic_runs'])) + ' problem|'
        check.status = 'WARN'
    return check


@action_function(start_runs=True, patch_completed=True)
def atacseq_start(connection, **kwargs):
    """Start runs by sending compiled input_json to run_workflow endpoint"""
    start = datetime.utcnow()
    action = ActionResult(connection, 'atacseq_start')
    my_auth = connection.ff_keys
    my_env = connection.ff_env
    fs_env = connection.fs_env
    atacseq_check_result = action.get_associated_check_result(kwargs).get('full_output', {})
    missing_runs = []
    patch_meta = []
    if kwargs.get('start_runs'):
        missing_runs = atacseq_check_result.get('needs_runs')
    if kwargs.get('patch_completed'):
        patch_meta = atacseq_check_result.get('completed_runs')
    action = wfr_utils.start_tasks(missing_runs, patch_meta, action, my_auth, my_env, fs_env, start,  move_to_pc=True, runtype='atac')
    return action



from dcicutils import ff_utils

# Use confchecks to import decorators object and its methods for each check module
# rather than importing check_function, action_function, CheckResult, ActionResult
# individually - they're now part of class Decorators in foursight-core::decorators
# that requires initialization with foursight prefix.
from .helpers.confchecks import *


# generic CHECK function used to add a static headers to items of some search result
def find_items_for_header_processing(connection, check, header, add_search=None,
                                     remove_search=None, append=True):
    """
    (add_search) and remove them from others (remove_search).
    Args are:
    - connection (FS connection)
    - check (required; check object initialized by CheckResult)
    - headers @id (required)
    - add_search search query
    - remove_search search query
    Meant to be used for CHECKS
    """
    # sets the full_output of the check!
    check.full_output = {'static_section': header, 'to_add': {}, 'to_remove': {}}
    # this GET will fail if the static header does not exist
    header_res = ff_utils.get_metadata(header, key=connection.ff_keys)
    # add entries keyed by item uuid with value of the static headers
    if add_search:
        search_res_add = ff_utils.search_metadata(add_search, key=connection.ff_keys)
        for search_res in search_res_add:
            curr_headers = search_res.get('static_headers', [])
            # handle case where frame != object
            if curr_headers and isinstance(curr_headers[0], dict):
                curr_headers = [obj['@id'] for obj in curr_headers]
            if header not in curr_headers:
                curr_headers = curr_headers + [header] if append else [header] + curr_headers
                check.full_output['to_add'][search_res['@id']] = curr_headers

    if remove_search:
        search_res_remove = ff_utils.search_metadata(remove_search,
                                                     key=connection.ff_keys)
        for search_res in search_res_remove:
            curr_headers = search_res.get('static_headers', [])
            # handle case where frame != object
            if curr_headers and isinstance(curr_headers[0], dict):
                curr_headers = [obj['@id'] for obj in curr_headers]
            if header in curr_headers:
                curr_headers.remove(header)
                check.full_output['to_remove'][search_res['@id']] = curr_headers

    if check.full_output['to_add'] or check.full_output['to_remove']:
        check.status = 'WARN'
        check.summary = 'Ready to add and/or remove static header'
        check.description = 'Ready to add and/or remove static header: %s' % header
        check.allow_action = True
        check.action_message = 'Will add static header to %s items and remove it from %s items' % (len(check.full_output['to_add']), len(check.full_output['to_remove']))
    else:
        check.status = 'PASS'
        check.summary = 'Static header is all set'


# generic ACTION function used along to add/remove static headers from the
# information obtained from find_items_for_header_processing
def patch_items_with_headers(connection, action, kwargs):
    """
    Arguments are:
    - the connection (FS connection)
    - the action (from ActionResult)
    - kwargs (from the action function)
    Takes care of patching info on Fourfront and also populating fields on the
    action
    """
    action_logs = {'patch_failure': [], 'patch_success': []}
    # get latest results from prepare_static_headers
    headers_check_result = action.get_associated_check_result(kwargs)
    # the dictionaries can be combined
    total_patches = headers_check_result['full_output']['to_add']
    total_patches.update(headers_check_result['full_output']['to_remove'])
    for item, headers in total_patches.items():
        # if all headers are deleted, use ff_utils.delete_field
        if headers == []:
            try:
                ff_utils.delete_field(item, 'static_headers', key=connection.ff_keys)
            except Exception as e:
                patch_error = '\n'.join([item, str(e)])
                action_logs['patch_failure'].append(patch_error)
            else:
                action_logs['patch_success'].append(item)
        else:
            patch_data = {'static_headers': headers}
            try:
                ff_utils.patch_metadata(patch_data, obj_id=item, key=connection.ff_keys)
            except Exception as e:
                patch_error = '\n'.join([item, str(e)])
                action_logs['patch_failure'].append(patch_error)
            else:
                action_logs['patch_success'].append(item)
    action.status = 'DONE'
    action.output = action_logs


@check_function(
    add_search='',
    remove_search='',
    header_at_id='',
    append=True,
    action="patch_static_headers"
)
def prepare_static_headers(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers')
    check.action = 'patch_static_headers'
    find_items_for_header_processing(connection, check, kwargs['header_at_id'],
                                     kwargs['add_search'], kwargs['remove_search'], append=kwargs['append'])
    return check


@action_function()
def patch_static_headers(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# Data Use Guidelines
@check_function(action="patch_static_headers_data_use_guidelines")
def prepare_static_headers_data_use_guidelines(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_data_use_guidelines')
    # only check experiment sets that are released, released to project, or in pre-release
    add_search = ('/search/?type=ExperimentSet&award.project=4DN&produced_in_pub.display_title=No%20value'
                  '&status=released&status=released+to+project&status=pre-release&frame=object')
    remove_search = ('/search/?type=ExperimentSet&award.project=4DN&produced_in_pub.display_title!=No%20value'
                     '&status=released&status=released+to+project&status=pre-release&frame=object')
    header_at_id = '/static-sections/621e8359-3885-40ce-965d-91894aa7b758/'
    check.action = 'patch_static_headers_data_use_guidelines'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search, append=False)
    return check


@action_function()
def patch_static_headers_data_use_guidelines(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_data_use_guidelines')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# InSitu Hi-C experiment Sets
@check_function(action="patch_static_headers_inSitu_HiC")
def prepare_static_headers_inSitu_HiC(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_inSitu_HiC')
    add_search = '/search/?experiments_in_set.experiment_type=in+situ+Hi-C&type=ExperimentSet&frame=object'
    remove_search = '/search/?experiments_in_set.experiment_type!=in+situ+Hi-C&type=ExperimentSet&frame=object'
    header_at_id = '/static-sections/298554ad-20e2-4449-a752-ac190123dab7/'
    check.action = 'patch_static_headers_inSitu_HiC'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search)
    return check


@action_function()
def patch_static_headers_inSitu_HiC(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_inSitu_HiC')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# Dilution Hi-C Experiment Sets
@check_function(action="patch_static_headers_dilution_HiC")
def prepare_static_headers_dilution_HiC(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_dilution_HiC')
    add_search = '/search/?experiments_in_set.experiment_type=Dilution+Hi-C&type=ExperimentSet&frame=object'
    remove_search = '/search/?experiments_in_set.experiment_type!=Dilution+Hi-C&type=ExperimentSet&frame=object'
    header_at_id = '/static-sections/7627f4eb-9f2d-4171-9e9b-87ab800ab5cd/'
    check.action = 'patch_static_headers_dilution_HiC'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search)
    return check


@action_function()
def patch_static_headers_dilution_HiC(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_dilution_HiC')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# FISH
@check_function(action="patch_static_headers_FISH")
def prepare_static_headers_FISH(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_FISH')
    add_search = '/search/?experiments_in_set.experiment_type=DNA+FISH&type=ExperimentSet&frame=object'
    remove_search = '/search/?experiments_in_set.experiment_type!=DNA+FISH&type=ExperimentSet&frame=object'
    header_at_id = '/static-sections/911424f9-21c7-49fc-b1df-865dd64ae91e/'
    check.action = 'patch_static_headers_FISH'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search)
    return check


@action_function()
def patch_static_headers_FISH(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_FISH')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# SPT
@check_function(action="patch_static_headers_SPT")
def prepare_static_headers_SPT(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_SPT')
    add_search = '/search/?experiments_in_set.experiment_type=SPT&type=ExperimentSet&frame=object'
    remove_search = '/search/?experiments_in_set.experiment_type!=SPT&type=ExperimentSet&frame=object'
    header_at_id = '/static-sections/6a313162-e70c-4fbe-93c5-bc78f5faf0c7/'
    check.action = 'patch_static_headers_SPT'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search)
    return check


@action_function()
def patch_static_headers_SPT(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_SPT')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# SPRITE
@check_function(action="patch_static_headers_SPRITE")
def prepare_static_headers_SPRITE(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_SPRITE')
    add_search = '/search/?experiments_in_set.experiment_type=DNA+SPRITE&type=ExperimentSet&frame=object'
    remove_search = '/search/?experiments_in_set.experiment_type!=DNA+SPRITE&type=ExperimentSet&frame=object'
    header_at_id = '/static-sections/205f35ec-92cd-4c02-bd35-b0d38dd72a90/'
    check.action = 'patch_static_headers_SPRITE'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search)
    return check


@action_function()
def patch_static_headers_SPRITE(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_SPRITE')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# MARGI
@check_function(action="patch_static_headers_MARGI")
def prepare_static_headers_MARGI(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_MARGI')
    add_search = '/search/?experiments_in_set.experiment_type=MARGI&type=ExperimentSet&frame=object'
    remove_search = '/search/?experiments_in_set.experiment_type!=MARGI&type=ExperimentSet&frame=object'
    header_at_id = '/static-sections/0c2ba23e-b256-47ce-a37c-0f1282471789/'
    check.action = 'patch_static_headers_MARGI'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search)
    return check


@action_function()
def patch_static_headers_MARGI(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_MARGI')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# sci-HiC
@check_function(action="patch_static_headers_sciHiC")
def prepare_static_headers_sciHiC(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_sciHiC')
    add_search = '/search/?experiments_in_set.experiment_type=sci-Hi-C&type=ExperimentSet&frame=object'
    remove_search = '/search/?experiments_in_set.experiment_type!=sci-Hi-C&type=ExperimentSet&frame=object'
    header_at_id = '/static-sections/ae5a6470-0694-4ba3-893a-40b170401bc0/'
    check.action = 'patch_static_headers_sciHiC'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search)
    return check


@action_function()
def patch_static_headers_sciHiC(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_sciHiC')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action


# DNase Hi-C
@check_function(action="patch_static_headers_DNase_HiC")
def prepare_static_headers_DNase_HiC(connection, **kwargs):
    check = CheckResult(connection, 'prepare_static_headers_DNase_HiC')
    add_search = '/search/?experiments_in_set.experiment_type=DNase+Hi-C&type=ExperimentSet&frame=object'
    remove_search = '/search/?experiments_in_set.experiment_type!=DNase+Hi-C&type=ExperimentSet&frame=object'
    header_at_id = '/static-sections/84448fd6-ccf0-45a7-86c8-673b5686c059/'
    check.action = 'patch_static_headers_DNase_HiC'
    find_items_for_header_processing(connection, check, header_at_id,
                                     add_search, remove_search)
    return check


@action_function()
def patch_static_headers_DNase_HiC(connection, **kwargs):
    action = ActionResult(connection, 'patch_static_headers_DNase_HiC')
    # get latest results from prepare_static_headers
    patch_items_with_headers(connection, action, kwargs)
    return action

#Chromatin Tracing
@check_function(action="patch_static_headers_Chromatin_Tracing")
def prepare_static_headers_Chromatin_Tracing(connection, **kwargs):
     check = CheckResult(connection, 'prepare_static_headers_Chromatin_Tracing')
     add_search = '/search/?experiments_in_set.experiment_type.display_title=multiplexed+FISH&type=ExperimentSet&frame=object'
     remove_search = '/search/?type=ExperimentSet&experiments_in_set.experiment_type.display_title%21=multiplexed+FISH&frame=object'
     header_at_id = '/static-sections/a09a2833-b56b-4e81-8eff-bb8ae6aaa596/'
     check.action = 'patch_static_headers_Chromatin_Tracing'
     find_items_for_header_processing(connection, check, header_at_id,
                                      add_search, remove_search)
     return check


@action_function()
def patch_static_headers_Chromatin_Tracing(connection, **kwargs):
     action = ActionResult(connection, 'patch_static_headers_Chromatin_Tracing')
     # get latest results from prepare_static_headers
     patch_items_with_headers(connection, action, kwargs)
     return action



import os
from foursight_core.checks.helpers.sys_utils import (
    parse_datetime_to_utc,
)
from dcicutils import ff_utils
from .helpers.google_utils import GoogleAPISyncer
import copy
import itertools
import datetime

# Use confchecks to import decorators object and its methods for each check module
# rather than importing check_function, action_function, CheckResult, ActionResult
# individually - they're now part of class Decorators in foursight-core::decorators
# that requires initialization with foursight prefix.
from .helpers.confchecks import *


#### HELPER FUNCTIONS ####

# cache results from requests to FF
ITEM_CACHE = {'titles': {}, 'replacing': {}}

def extract_info(res, fields):
    return {field: res[field] for field in fields if field in res}


def extract_list_info(res_list, fields, key_field):
    if not res_list:
        return {}
    results = {}
    for res in res_list:
        if key_field not in res:
            continue
        results[res[key_field]] = extract_info(res, fields)
    return results


def add_to_report(exp_set, exp_sets):
    """
    Used to process search hits in experiment_set_reporting_data
    """
    exp_set_res = extract_info(exp_set, ['@id', 'status', 'tags'])
    exp_set_res['award.project'] = exp_set.get('award', {}).get('project')
    exp_set_res['processed_files'] = extract_list_info(
        exp_set.get('processed_files'),
        ['@id', 'status', 'file_type'],
        'accession'
    )
    exps = {}
    for exp in exp_set.get('experiments_in_set', []):
        exp_res = extract_info(exp, ['@id', 'status', 'experiment_type'])
        exp_res['files'] = extract_list_info(
            exp.get('files'),
            ['@id', 'status', 'file_type'],
            'accession'
        )
        exp_res['processed_files'] = extract_list_info(
            exp.get('processed_files'),
            ['@id', 'status', 'file_type'],
            'accession'
        )
        exps[exp['accession']] = exp_res
    exp_set_res['experiments_in_set'] = exps
    exp_sets[exp_set['accession']] = exp_set_res


def find_item_title(item_id):
    """
    Data is always used for release updates, so hardcode here (dirty hack)
    Use cache to improve performance
    Return display title of given item, None if item can't be found
    """
    if item_id == 'UNKNOWN_ID':
        return None
    if item_id in ITEM_CACHE['titles']:
        return ITEM_CACHE['titles'][item_id]
    item_obj = ff_utils.get_metadata(item_id, ff_env='data', add_on='frame=object')
    title = item_obj.get('display_title')
    ITEM_CACHE['titles'][item_id] = title
    return title


def find_replacing_item(item_id):
    """
    Data is always used for release updates, so hardcode here (dirty hack)
    Use cache to improve performance
    Return the @id of replacing item, None if item can't be found
    """
    if item_id == 'UNKNOWN_ID':
        return None
    if item_id in ITEM_CACHE['replacing']:
        return ITEM_CACHE['replacing'][item_id]
    # need to get accession manually from replaced item to find the replacing item
    replaced = ff_utils.get_metadata(item_id, ff_env='data', add_on='frame=object')
    replaced_accession = replaced['accession']
    replacing = ff_utils.get_metadata(replaced_accession, ff_env='data', add_on='frame=object')
    if not replacing.get('alternate_accessions'):  # see if it is actually a replacing item
        replacing = None
    else:
        replacing = replacing.get('@id')
    ITEM_CACHE['replacing'][item_id] = replacing
    return replacing


def calculate_report_from_change(path, prev, curr, add_ons):
    exp_type = add_ons.get('exp_type', 'UNKNOWN_TYPE')
    file_type = add_ons.get('file_type', 'UNKNOWN_TYPE')
    item_id = add_ons.get('@id', 'UNKNOWN_ID')
    set_id = add_ons.get('set_@id', 'UNKNOWN_ID')
    released_exp_set = add_ons.get('released_exp_set', False)
    released_exp = add_ons.get('released_exp', False)
    equivalents = add_ons.get('equivalents', {})
    significants = add_ons.get('significants', {})
    field = path.split('.')[-1]
    # these dictionaries define the reports
    # *NUM* is replaced by the number of reports
    # to save time, calculate conditional functions within report_info here
    # calculate info for REPLACED items
    if curr in significants[field] and curr == 'replaced':
        replacing_secondary_id = find_replacing_item(item_id)
        if item_id == set_id:  # we are on an exp set
            replacing_add_info = 'Replaced by'
        else:
            replacing_add_info = '%s replaced by' % find_item_title(item_id)
    else:
        replacing_secondary_id = None
        replacing_add_info = ''

    report_info_w_released_exp_set = {
        'experiments_in_set.status': {
            '*/released' : {
                'severity': 1,
                'priority': 3,
                'summary': 'New replicate experiments have been added to a released %s replicate set.' % exp_type,
                'summary_plural': 'New replicate experiments have been added to *NUM* released %s replicate sets.' % exp_type,
            },
            '*/*': {
                'severity': 3,
                'priority': 4,
                'summary': 'Replicate experiments with status %s have been added to a released %s replicate set.' % (curr, exp_type),
                'summary_plural': 'Replicate experiments with status %s have been added to *NUM* released %s replicate sets.' % (curr, exp_type),
            },
        },
        'processed_files.status': {
            '*/released': {
                'severity': 0,
                'priority': 7,
                'summary': 'New processed files have been added to a released %s replicate set.' % exp_type,
                'summary_plural': 'New processed files have been added to *NUM* released %s replicate sets.' % exp_type,
            }
        }
    }
    report_info_w_released_exp = {
        'experiments_in_set.files.status': {
            '*/released': {
                'severity': 2,
                'priority': 5,
                'summary': 'New raw files have been added to a released %s replicate set.' % exp_type,
                'summary_plural': 'New raw files have been added to *NUM* released %s replicate sets.' % exp_type,
            }
        },
        'experiments_in_set.files.status': {
            '*/*': {
                'severity': 3,
                'priority': 6,
                'summary': 'New unreleased raw files have been added to a released %s replicate set.' % exp_type,
                'summary_plural': 'New unreleased raw files have been added to *NUM* released %s replicate sets.' % exp_type,
            }
        },
        'experiments_in_set.processed_files.status': {
            '*/released': {
                'severity': 0,
                'priority': 7,
                'summary': 'New processed files have been added to a released %s replicate set.' % exp_type,
                'summary_plural': 'New processed files have been added to *NUM* released %s replicate sets.' % exp_type,
            }
        }
    }
    report_info = {
        'status': {
            '*/released' : {
                'severity': 0,
                'priority': 0, # 0 is highest priority
                'summary': 'New %s replicate set has been released.' % exp_type,
                'summary_plural': '*NUM* new %s replicate sets have been released.' % exp_type,
            },
            'archived/released' : {
                'severity': 3,
                'priority': 0,
                'summary': 'New %s replicate set has changed from archived to released.' % exp_type,
                'summary_plural': '*NUM* new %s replicate sets have changed from archived to released.' % exp_type,
            },
            'replaced/released' : {
                'severity': 3,
                'priority': 0,
                'summary': 'New %s replicate set has changed from replaced to released.' % exp_type,
                'summary_plural': '*NUM* new %s replicate sets have changed from replaced to released.' % exp_type,
            },
            'released/archived' : {
                'severity': 1,
                'priority': 1,
                'summary': 'Released %s replicate set has been archived.' % exp_type,
                'summary_plural': '*NUM* released %s replicate sets have been archived.' % exp_type,
            },
            'released/replaced' : {
                'severity': 1,
                'priority': 1,
                'summary': 'Released %s replicate set has been replaced.' % exp_type,
                'summary_plural': '*NUM* released %s replicate sets have been replaced.' % exp_type,
                'secondary_id': replacing_secondary_id,
                'add_info': replacing_add_info
            },
            'released/*': {
                'severity': 3,
                'priority': 2,
                'summary': 'Released %s replicate set has changed to %s.' % (exp_type, curr),
                'summary_plural': '*NUM* released %s replicate sets have changed to %s.' % (exp_type, curr),
            }
        },
        'experiments_in_set.status': {
            'released/archived' : {
                'severity': 1,
                'priority': 1,
                'summary': 'Released replicate experiments from a %s replicate set have been archived.' % exp_type,
                'summary_plural': 'Released replicate experiments from *NUM* %s replicate sets have been archived.' % exp_type,
            },
            'released/replaced': {
                'severity': 1,
                'priority': 1,
                'summary': 'Released replicate experiments from a %s replicate set have been replaced.' % exp_type,
                'summary_plural': 'Released replicate experiments from *NUM* %s replicate sets have been replaced.' % exp_type,
                'secondary_id': replacing_secondary_id,
                'add_info': replacing_add_info
            },
            'released/*': {
                'severity': 3,
                'priority': 2,
                'summary': 'Replicate experiments from a %s replicate set have changed from released to %s.' % (exp_type, curr),
                'summary_plural': 'Replicate experiments from *NUM* %s replicate sets have changed from released to %s.' % (exp_type, curr),
            },
        },
        'processed_files.status': {
            'released/archived': {
                'severity': 1,
                'priority': 8,
                'summary': 'Released processed files from a %s replicate set have been archived.' % exp_type,
                'summary_plural': 'Released processed files from *NUM* %s replicate sets have been archived.' % exp_type,
            },
            'released/replaced': {
                'severity': 1,
                'priority': 8,
                'summary': 'Released processed files from a %s replicate set have been replaced.' % exp_type,
                'summary_plural': 'Released processed files from *NUM* %s replicate sets have been replaced.' % exp_type,
                'secondary_id': replacing_secondary_id,
                'add_info': replacing_add_info
            },
            'released/*': {
                'severity': 3,
                'priority': 9,
                'summary': 'Processed files from a %s replicate set have been changed from released to %s.' % (exp_type, curr),
                'summary_plural': 'Processed files from *NUM* %s replicate sets have been changed from released to %s.' % (exp_type, curr),
            }
        },
        'experiment_in_set.processed_files.status': {
            'released/archived': {
                'severity': 1,
                'priority': 8,
                'summary': 'Released processed files from a %s replicate set have been archived.' % exp_type,
                'summary_plural': 'Released processed files from *NUM* %s replicate sets have been archived.' % exp_type,
            },
            'released/replaced': {
                'severity': 1,
                'priority': 8,
                'summary': 'Released processed files from a %s replicate set have been replaced.' % exp_type,
                'summary_plural': 'Released processed files from *NUM* %s replicate sets have been replaced.' % exp_type,
                'secondary_id': replacing_secondary_id,
                'add_info': replacing_add_info
            },
            'released/*': {
                'severity': 3,
                'priority': 9,
                'summary': 'Processed files from a %s replicate set have been changed from released to %s.' % (exp_type, curr),
                'summary_plural': 'Processed files from *NUM* %s replicate sets have been changed from released to %s.' % (exp_type, curr),
            }
        },
        'experiments_in_set.files.status': {
            'released/archived': {
                'severity': 1,
                'priority': 8,
                'summary': 'Released raw files from a %s replicate set have been archived.' % exp_type,
                'summary_plural': 'Released raw files from *NUM* %s replicate sets have been archived.' % exp_type,
            },
            'released/replaced': {
                'severity': 1,
                'priority': 8,
                'summary': 'Released raw files from a %s replicate set have been replaced.' % exp_type,
                'summary_plural': 'Released raw files from *NUM* %s replicate sets have been replaced.' % exp_type,
                'secondary_id': replacing_secondary_id,
                'add_info': replacing_add_info
            },
            'released/*': {
                'severity': 3,
                'priority': 9,
                'summary': 'Raw files from a %s replicate set have changed from released to %s.' % (exp_type, curr),
                'summary_plural': 'Raw files from *NUM* %s replicate sets have changed from released to %s.' % (exp_type, curr),
            }
        }
    }
    # update the reporting framework if necessary
    if released_exp_set:
        report_info.update(report_info_w_released_exp_set)
    if released_exp:
        report_info.update(report_info_w_released_exp)

    if field in significants:
        # get the equivalent value for the field if it matches
        if field in equivalents:
            prev = equivalents[field].get(prev, prev)
            curr = equivalents[field].get(curr, curr)
        # if the field does not have a significant value, set to '*'
        prev_key = prev if prev in significants[field] else '*'
        curr_key = curr if curr in significants[field] else '*'
        level_1 = report_info.get(path)
        if level_1:
            level_2 = level_1.get('/'.join([prev_key, curr_key]))
            if level_2:
                secondary_id = level_2.get('secondary_id')
                level_2['items'] = [{'secondary_id': secondary_id if secondary_id else item_id,
                                     'additional_info': level_2.get('add_info', '')}]
                level_2['set_@id'] = set_id
                # clean up unneeded fields
                if 'secondary_id' in level_2:
                    del level_2['secondary_id']
                if 'add_info' in level_2:
                    del level_2['add_info']
                return level_2
    # no report found
    return None


def generate_exp_set_report(curr_res, prev_res, **kwargs):
    """
    curr_res and prev_res are dictionary objects to compare.
    kwargs should include:
    - add_ons (dict important info passed to calculate_report_from_change)
    - field_path (list representing the path of objects we have traversed)
    - report_fields (list of which fields are significant to report on)
    - released_statuses (list of which statuses to consider "released")
        Maybe this should be deprecated in favor of using existing equivalents in add_ons...
    - children_fields (list of which fields could be child objects)
    """
    reports = []
    # kwargs
    add_ons = copy.copy(kwargs.get('add_ons', {}))
    field_path = kwargs.get('field_path', [])
    report_fields = kwargs.get('report_fields', [])
    released_statuses = kwargs.get('released_statuses', [])
    children_fields = kwargs.get('children_fields', [])
    # set some top level exp set attributes in add_ons
    # top level because all exp types should be identical in a replicate set
    if field_path == []:
        exps_in_set = curr_res.get('experiments_in_set')
        if exps_in_set:
            first_exp_type = exps_in_set[list(exps_in_set.keys())[0]].get('experiment_type')
            if first_exp_type:
                add_ons['exp_type'] = first_exp_type
        add_ons['set_@id'] = curr_res.get('@id')
    # handle file_type add_ons. cannot be both processed_files and files
    # currently this add_on isn't used for anything
    if field_path and field_path[-1] in ['processed_files', 'files']:
        file_type = curr_res.get('file_type')
        if file_type:
            add_ons['file_type'] = file_type
    for key, val in curr_res.items():
        path = '.'.join(field_path + [key])
        curr_val = curr_res.get(key, None)
        prev_val = prev_res.get(key, None)
        # catch cases where certain fields are missing, which is probably a schema adjustment issue
        if (curr_res and curr_val is None) or (prev_res and prev_val is None):
            continue
        # START add_ons
        if path == 'status' and curr_val in released_statuses and prev_val in released_statuses:
            add_ons['released_exp_set'] = True
        if path == 'experiments_in_set.status' and curr_val in released_statuses and prev_val in released_statuses:
            add_ons['released_exp'] = True
        # END add_ons
        if key in children_fields and isinstance(val, dict):
            for c_key, c_val in val.items():
                if isinstance(c_val, dict) and '@id' in c_val:
                    prev_child = prev_res.get(key, {}).get(c_key, {})
                    child_path = field_path + [key]
                    child_reports = generate_exp_set_report(
                        c_val,
                        prev_child,
                        report_fields=report_fields,
                        released_statuses=released_statuses,
                        children_fields=children_fields,
                        field_path=child_path,
                        add_ons=add_ons
                    )
                    if child_reports:
                        reports.append(child_reports)
        elif key in report_fields:
            add_ons['@id'] = curr_res['@id']
            report = calculate_report_from_change(path, prev_val, curr_val, add_ons)
            if report:
                reports.append(report)
    # merge reports with the same summary to combine @ids into 'items' field
    merging = {}
    for report in reports:
        if report['summary'] not in merging:
            merging[report['summary']] = report
        else:
            merging[report['summary']]['items'].extend(report['items'])
    merged_reports = list(merging.values())
    if merged_reports:
        merged_reports.sort(key = lambda r: r['priority'])
    return merged_reports[0] if merged_reports else None


#### CHECKS / ACTIONS #####

# TODO: This check has been removed from the schedule and should be revisited and refactored
# @check_function()
def experiment_set_reporting_data(connection, **kwargs):
    """
    Get a snapshot of all experiment sets, their experiments, and files of
    all of the above. Include uuid, accession, status, and md5sum (for files).
    """
    check = CheckResult(connection, 'experiment_set_reporting_data')
    check.status = 'IGNORE'
    exp_sets = {}
    search_query = '/search/?type=ExperimentSetReplicate&experimentset_type=replicate&sort=-date_created'
    set_hits = ff_utils.search_metadata(search_query, key=connection.ff_keys, page_limit=20)
    # run a second search for status=deleted and status=replaced
    set_hits_del = ff_utils.search_metadata(search_query + '&status=deleted&status=replaced',
                                            key=connection.ff_keys, page_limit=20)
    set_hits.extend(set_hits_del)
    for hit in set_hits:
        add_to_report(hit, exp_sets)
    check.full_output = exp_sets
    return check


def _old_data_release_updates(connection, **kwargs):
    """
    OLD VERSION of the data_release_updates check with docstring.

    Diff two results of 'experiment_set_reporting_data' check.

    start_date and end_date are dates in form YYYY-MM-DD.
    Check will find experiment_set_reporting_data results closest to 12pm UTC
    for both those dates. If None is provided, it will consider it an empty
    experiment_set_reporting_data result for start_date; this may
    be used to effectively generate reports from a clean slate.

    If None is provided to the end date, the primary result of
    experiment_set_reporting_data is used.

    update_tag is the tag used to organize updates in Fourfront.

    tag_filter is a string value required to be in the 'tags' field of a given
    replicate set for the set to be considered by this check. It defaults to
    None, which means no tag is required. Another likely value is
    '4DN Joint Analysis 2018'.

    project_filter is a string value that is required to match the award.project
    field of the replicate set to be considered by this check. If set to None,
    all replicate sets will be used regardless of project.

    Stores the information used by that action to actually build reports.
    """
    check = CheckResult(connection, 'data_release_updates')
    check.action = 'publish_data_release_updates'
    # find needed experiment_set_reporting_data results
    data_check = CheckResult(connection, 'experiment_set_reporting_data')
    if kwargs.get('start_date', None) is not None:
        start_date = parse_datetime_to_utc(kwargs.get('start_date'), '%Y-%m-%d')
        # use 11 am UTC
        start_date = start_date.replace(hour=11)
        start_date_str = datetime.datetime.strftime(start_date, '%Y-%m-%d')
        start_data_result = data_check.get_closest_result(override_date=start_date)
    else:
        start_data_result = {}
        start_date_str = '2017-01-01'  # Arbitrary earliest time to display

    if kwargs.get('end_date'):
        end_date = parse_datetime_to_utc(kwargs.get('end_date'), '%Y-%m-%d')
        # use 11 am UTC by default
        end_date = end_date.replace(hour=11)
        end_date_str = datetime.datetime.strftime(end_date, '%Y-%m-%d')
        end_data_result = data_check.get_closest_result(override_date=end_date)
    else:  # use primary result if end_data_result is not supplied
        end_data_result = data_check.get_primary_result()
        end_date_str = end_data_result['uuid'][:10]  # YYYY-MM-DD
    # start_date must be before end_date
    if start_date_str > end_date_str:
        check.status = 'ERROR'
        check.description = 'start_date cannot be greater than end_date.'
        return check
    if (kwargs.get('start_date') and not start_data_result) or not end_data_result:
        check.status = 'ERROR'
        check.description = 'One or both experiment_set_reporting_data results are not available.'
        return check
    tag_filter = kwargs['tag_filter']
    project_filter = kwargs.get('project_filter')
    used_res_strings = ' Compared results from %s (start) to %s (end). UUIDs are %s (start) and %s (end). Used filter on replicate sets tag is %s and the filter on award.project is %s (None means no filter in both cases).' % (
    start_date_str, end_date_str, start_data_result.get('uuid', 'None'), end_data_result['uuid'], tag_filter,
    project_filter)
    start_output = start_data_result.get('full_output', {})  # this can be empty
    end_output = end_data_result.get('full_output')  # this cannot be empty
    if not isinstance(start_output, dict) or not isinstance(end_output, dict):
        check.status = 'ERROR'
        check.description = 'One or both experiment_set_reporting_data results are malformed.' + used_res_strings
        return check

    reports = []
    ### INFO USED TO CONTROL REPORTING
    report_fields = ['status']  # fields we care about reporting on
    children_fields = ['experiments_in_set', 'files', 'processed_files']  # possible linkTo fields
    add_ons = {}
    add_ons['significants'] = {
        'status': ['released', 'archived', 'replaced']
    }

    ### THIS DIFFERENTIATES PUBLIC VS PUBLIC + INTERNAL RELEASE UPDATES
    if kwargs['is_internal'] == True:
        # effectively consider released and released_to_project the same
        released_statuses = ['released', 'released to project']
        add_ons['equivalents'] = {
            'status': {
                'released to project': 'released',
                'archived_to_project': 'archived'
            }
        }
    else:
        released_statuses = ['released']

    ### CREATE REPORTS... assuming experiment sets will NOT be deleted from DB
    for exp_set in end_output:
        end_res = end_output[exp_set]
        # apply filters
        if tag_filter and tag_filter not in end_res.get('tags', []):
            continue
        if project_filter and project_filter != end_res.get('award.project'):
            continue
        start_res = start_output.get(exp_set, {})
        exp_set_report = generate_exp_set_report(
            end_res,
            start_res,
            report_fields=report_fields,
            released_statuses=released_statuses,
            children_fields=children_fields,
            add_ons=add_ons
        )
        if exp_set_report is not None:
            reports.append(exp_set_report)

    ### GROUP UPDATES BY SUMMARY AND BUILD METADATA
    reports.sort(key=lambda r: r['summary'])
    check.full_output = reports
    group_reports = []
    # group reports by summary
    for key, group in itertools.groupby(reports, lambda r: r['summary']):
        # first item
        first_report = group.__next__()
        group_report = {r_key: first_report[r_key] for r_key in ['severity', 'summary', 'summary_plural']}
        first_report_items = {'primary_id': first_report['set_@id'], 'secondary_ids': first_report['items']}
        group_report['update_items'] = [first_report_items]
        # use this checks uuid as the timestamp for the reports
        group_report['foursight_uuid'] = kwargs['uuid'] + '+00:00'
        group_report['update_tag'] = kwargs['update_tag']
        group_report['end_date'] = end_date_str
        group_report['start_date'] = start_date_str
        # handle parameters like tag_filter and project_filter used to create report
        group_report['parameters'] = []
        if tag_filter:
            group_report['parameters'].append('tags=' + tag_filter)
        if project_filter:
            group_report['parameters'].append('award.project=' + project_filter)
        # use 4DN DCIC lab and award, with released status
        group_report['lab'] = '4dn-dcic-lab'
        group_report['award'] = '1U01CA200059-01'
        group_report['status'] = 'released'
        group_report['is_internal'] = kwargs['is_internal']
        for gr in group:  # iterate through the rest of the items
            group_report['update_items'].append({'primary_id': gr['set_@id'], 'secondary_ids': gr['items']})
        group_reports.append(group_report)
    # update summaries for plural cases
    for group_report in group_reports:
        num_items = len(group_report['update_items'])
        if num_items > 1:
            group_report['summary_plural'] = group_report['summary_plural'].replace('*NUM*', str(num_items))
            group_report['summary'] = group_report['summary_plural']
        del group_report['summary_plural']
    # lastly make the static section for this update_tag
    static_proj = project_filter if project_filter else ''
    static_scope = 'network members' if kwargs['is_internal'] else 'public'
    static_tag = 'with %s tag' % tag_filter if tag_filter else ''
    static_content = ' '.join(
        ['All', static_proj, 'data released to', static_scope, 'between', start_date_str, 'and', end_date_str,
         static_tag])
    report_static_section = {
        'name': 'release-updates.' + kwargs['update_tag'],
        'body': '<h4 style=\"margin-top:0px;font-weight:400\">' + static_content + '</h4>'
    }
    check.status = 'PASS'
    if group_reports:
        check.brief_output = {
            'release_updates': group_reports,
            'static_section': report_static_section
        }
        check.description = 'Ready to publish new data release updates.' + used_res_strings
        check.action_message = 'Will publish %s  grouped updates with the update_tag: %s. See brief_output.' % (
        str(len(group_reports)), kwargs['update_tag'])
        check.allow_action = True
    else:
        check.brief_output = {
            'release_updates': [],
            'static_section': None
        }
        check.description = 'There are no data release updates for the given dates and filters.' + used_res_strings
    return check


# TODO: This check has been removed from the schedule and should be revisited and refactored
# @check_function(start_date=None, end_date=None, update_tag='DISCARD', tag_filter=None, project_filter='4DN', is_internal=False)
def data_release_updates(connection, **kwargs):
    """ TODO: New version of this check - for now, does nothing - see old version above. """
    check = CheckResult(connection, 'data_release_updates')
    check.status = 'PASS'
    check.brief_output = check.full_output = 'This check still needs to be implemented!'
    return check


# TODO: This action has been removed from the schedule and should be revisited and refactored
# @action_function
def publish_data_release_updates(connection, **kwargs):
    """ TODO: This action probably needs rewriting as well as it based on the OLD data_release_updates check. """
    action = ActionResult(connection, 'publish_data_release_updates')
    report_result = action.get_associated_check_result(kwargs)
    action.description = "Publish data release updates to Fourfront."
    updates_to_post = report_result.get('brief_output', {}).get('release_updates', [])
    section_to_post = report_result.get('brief_output', {}).get('static_section')
    # post items to FF
    posted_updates = []
    for update in updates_to_post:
        # should be in good shape to post as-is
        resp = ff_utils.post_metadata(update, 'data-release-updates', key=connection.ff_keys)
        posted_updates.append({'update': update, 'response': resp})
    if section_to_post:
        resp = ff_utils.post_metadata(section_to_post, 'static-sections', key=connection.ff_keys)
        posted_section = {'static_section': section_to_post, 'response': resp}
    else:
        posted_section = None
    action.output = {
        'updates_posted': posted_updates,
        'section_posted': posted_section
    }
    action.status = 'DONE'
    return action


@check_function(start_date=None, end_date=None)
def sync_google_analytics_data(connection, **kwargs):
    '''
    This checks the last time that analytics data was fetched (if any) and then
    triggers an action to fill up fourfront with incremented google_analytics TrackingItems.

    TODO: No use case yet, but we could accept start_date and end_date here & maybe in action eventually.
    '''
    check = CheckResult(connection, 'sync_google_analytics_data')

    if os.environ.get('chalice_stage', 'dev') != 'prod':
        check.summary = check.description = 'This check only runs on Foursight prod'
        return check

    recent_passing_run = False
    recent_runs, total_unused = check.get_result_history(0, 20, after_date=datetime.datetime.now() - datetime.timedelta(hours=3))
    for run in recent_runs:
        # recent_runs is a list of lists. [status, None, kwargdict]
        # Status is at index 0.
        if run[0] == 'PASS':
            recent_passing_run = True
            break

    if recent_passing_run:
        check.summary = check.description = 'This check was run within last 3 hours; skipping because need time for TrackingItems to be indexed.'
        check.status = 'FAIL'
        return check

    google = GoogleAPISyncer(connection.ff_keys)

    action_logs = { 'daily_created' : [], 'monthly_created' : [] }

    res = google.analytics.fill_with_tracking_items('daily')
    action_logs['daily_created'] = res.get('created', [])

    res = google.analytics.fill_with_tracking_items('monthly')
    action_logs['monthly_created'] = res.get('created', [])

    check.full_output = action_logs
    check.status = 'PASS' if (len(action_logs['daily_created']) > 0 or len(action_logs['monthly_created']) > 0) else 'WARN'
    check.description = 'Created %s daily items and %s monthly Items.' % (str(len(action_logs['daily_created'])), str(len(action_logs['monthly_created'])))
    return check
